import maude
import os.path

maude.init(advise=False)
maude.load(os.path.join(os.path.dirname(__file__), '..', 'example.maude'))

m = maude.getCurrentModule()

print('Using', m, 'module')

#####

t = m.parseTerm('f(a, a)')
v = m.parseTerm('H:Symbol')

for r, sb, ctx, rl in t.apply('ab'):
	print(r, 'with', rl, 'in context', ctx(v))

for r, sb, ctx, rl in t.apply('bc'):
	print(r, 'with', rl, 'in context', ctx(v))

#####

t = m.parseTerm('f(a, b)')

for r, sb, ctx, rl in t.apply('ab'):
	print(r, 'with', rl, 'in context', ctx(v))

	for s, sb, ctx, rl in r.apply('bc'):
		print(s, 'with', rl, 'in context', ctx(v))

for r, sb, ctx, rl in t.apply(None):
	print(r, 'with', rl, 'in context', ctx(v))

#####

for r, sb, ctx, rl in t.apply('swap'):
	print(r, 'with', rl, 'in context', ctx(v), 'and substitution', sb)

#####

t = m.parseTerm('f(f(a, g(a)), f(b, c))')
sb = maude.Substitution({m.parseTerm('X'): m.parseTerm('b')})

for r, *_ in t.apply('swap', minDepth=1):
	print(r)

for r, *_ in t.apply('swap', maxDepth=-1):
	print(r)

for r, *_ in t.apply('swap', substitution=sb):
	print(r)

#####

swap = next(rl for rl in m.getRules() if rl.getLabel() == 'swap')

def apply_rule(t, rl, maxDepth=maude.UNBOUNDED):
	"""Apply a rule"""

	return [ctx(match.instantiate(rl.getRhs())) for match, ctx in t.match(rl.getLhs(), maxDepth=maxDepth)]

print(apply_rule(t, swap))
print(apply_rule(t, swap, maxDepth=0))

#####

def apply_rule2(t, rl, substitution=None, maxDepth=maude.UNBOUNDED):
	"""Apply a rule with an optional initial substitution"""

	lhs = rl.getLhs()
	rhs = rl.getRhs()
	condition = list(rl.getCondition())

	if substitution:
		lhs = substitution.instantiate(lhs)
		rhs = substitution.instantiate(rhs)

		for k, cf in enumerate(condition):
			if isinstance(cf, maude.SortTestCondition):
				condition[k] = maude.SortTestCondition(substitution.instantiate(cf.getLhs()), cf.getSort())
			else:
				condition[k] = type(cf)(substitution.instantiate(cf.getLhs()),
				                        substitution.instantiate(cf.getRhs()))

	return [ctx(match.instantiate(rhs)) for match, ctx in t.match(lhs, condition=condition, maxDepth=maxDepth)]

print(apply_rule2(t, swap))
print(apply_rule2(t, swap, substitution=sb))




#
# Toy interactive rewriting interface using the maude library
# for "Maude as a library: an efficient all-purpose programming interface"
#

import cmd
import subprocess
import sys
import tempfile

import maude


def print_tree(term, indent=''):
	"""Print a term as a tree"""

	print(f'{indent}{term.symbol()} : {term.getSort()}')

	indent = indent + '  '

	for argument in term.arguments():
		print_tree(argument, indent)


def count_symbols(term, count):
	"""Count occurrences of symbols"""

	count[term.symbol()] = 1 + count.get(term.symbol(), 0)
	for argument in term.arguments():
		count_symbols(argument, count)


def find_vars(term, varset):
	"""Get all variables in term"""

	if term.isVariable():
		varset.add(term)
	else:
		for argument in term.arguments():
			find_vars(argument, varset)


def needs_term(func):
	"""Check that a term exists"""

	def func2(self, *args):
		if self.term:
			func(self, *args)
		else:
			print('Error: no current term.')

	# Copy function's documentation
	func2.__doc__ = func.__doc__

	return func2


def needs_module(func):
	"""Check that a term exists"""

	def func2(self, *args):
		if self.module:
			func(self, *args)
		else:
			print('Error: no current module.')

	# Copy function's documentation
	func2.__doc__ = func.__doc__

	return func2


class InteractiveRewriter(cmd.Cmd):
	"""Interactive rewriter for Maude"""

	# Introductory message
	intro = '\n     *** Interactive rewriter for Maude ***\n'
	prompt = 'IRew> '

	def __init__(self):
		super().__init__()

		# Current term
		self.term = None
		# Current module
		self.module = maude.getCurrentModule()
		# Term history
		self.history = []

		# Current metamodule (because we may add rules)
		self.metamodule = None

	def do_load(self, filename):
		"""Load a Maude file"""

		maude.load(filename)
		self.module = maude.getCurrentModule()

	def do_select(self, name):
		"""Select a module"""

		self.module = maude.getModule(name)
		self.term = None

	def do_list(self, _):
		"""List known modules"""

		for x in maude.getModules():
			print(x)

	@needs_module
	def do_start(self, text):
		"""Start with the given term"""

		self.term = self.module.parseTerm(text)

	@needs_term
	def do_show(self, _):
		"""Show the current term"""

		print('The current term is', self.term)

	@needs_term
	def do_tree(self, _):
		"""Show the current term as a tree"""

		print_tree(self.term)

	@needs_term
	def do_count(self, _):
		"""Show the number of occurrences of each symbol"""

		count = {}
		count_symbols(self.term, count)

		for symbol, times in count.items():
			print(f'{str(symbol):30} : {times}')

	@needs_term
	def do_reduce(self, _):
		"""Reduce the current term"""

		nrew = self.term.reduce()
		print(f'Reduced to {self.term} in {nrew} rewrites.')

	@needs_term
	def do_srewrite(self, text):
		"""Reduce the current term with a strategy"""

		strategy = self.module.parseStrategy(text)
		for result, nrew in self.term.srewrite(strategy):
			print(f'{result} in {nrew} rewrites')

	@needs_term
	def do_step(self, label):
		"""Take a rewriting step interactive"""

		results = []

		for k, (result, subs, ctx, rl) in enumerate(self.term.apply(label if label else None)):
			where = self.print_context(ctx, rl.getLhs())
			print(f'({k}) {result} by applying {rl} on {where} with {subs}')
			results.append(result)

		self.select_one(results)

	@needs_term
	def do_inline(self, text):
		"""Apply a inline rewrite rule"""

		arrow_index = text.index('=>')
		lhs = self.module.parseTerm(text[:arrow_index])
		rhs = self.module.parseTerm(text[arrow_index + 2:])

		results = []  # results of the inline rewriting

		for k, (subs, ctx) in enumerate(self.term.match(lhs, maxDepth=maude.UNBOUNDED)):
			result = ctx(subs.instantiate(rhs))
			where = self.print_context(ctx, lhs)

			print(f'({k}) {result} in {where} with {subs}')
			results.append(result)

		self.select_one(results)

	@needs_module
	def do_add(self, text):
		"""Add a rule to the current module"""

		arrow_index = text.index('=>')
		lhs = self.module.parseTerm(text[:arrow_index])
		rhs = self.module.parseTerm(text[arrow_index + 2:])

		ml = maude.getModule('META-LEVEL')

		if self.metamodule is None:
			self.metamodule = ml.parseTerm(f"upModule('{self.module}, false)")
			self.metamodule.reduce()

		# Find the metarule operator
		term_kind = ml.findSort('Term').kind()
		rule_kind = ml.findSort('Rule').kind()
		attr_kind = ml.findSort('Attr').kind()
		none_term = ml.parseTerm('none', attr_kind)

		rl_symb = ml.findSymbol('rl_=>_[_].', (term_kind, term_kind, attr_kind), rule_kind)

		# Generate the new rule at the metalevel
		rl_term = rl_symb.makeTerm((ml.upTerm(lhs), ml.upTerm(rhs), none_term))

		# Insert the rule into the module
		mm_args = list(self.metamodule.arguments())
		rls_symb = ml.findSymbol('__', (rule_kind, rule_kind), rule_kind)

		mm_args[7] = rls_symb(mm_args[7], rl_term)

		self.metamodule = self.metamodule.symbol().makeTerm(mm_args)

		# Update the current module
		self.module = maude.downModule(self.metamodule)

		if self.term:
			self.term = self.module.parseTerm(str(self.term))

		print('The rule has been inserted.')

	@needs_module
	def do_trs(self, _):
		"""Generate a TRS file from the rules of the current module"""

		self.make_trs(sys.stdout)

	@needs_module
	def do_termination(self, _):
		"""Check the termination of the rules in the current module"""

		with tempfile.NamedTemporaryFile('w', suffix='.trs') as tmpf:
			self.make_trs(tmpf)
			tmpf.flush()
			subprocess.run(('java', '-jar', 'aprove.jar', '-m', 'wst', tmpf.name, '-p', 'plain'))

	@needs_module
	def do_confluence(self, _):
		"""Check confluence of the rules in the current module"""

		with tempfile.NamedTemporaryFile('w', suffix='.trs') as tmpf:
			self.make_trs(tmpf)
			tmpf.flush()

			try:
				subprocess.run(('csi.sh', tmpf.name))

			except FileNotFoundError:
				print('Cannot find csi.sh from the CSI tool in the system path.\n'
				      'It can downloaded from http://cl-informatik.uibk.ac.at/software/csi/.')

	def do_EOF(self, _):
		"""Exits the interpreter"""

		print()
		return True

	def select_one(self, options):
		"""Select one of the options for the current term"""

		if not options:
			print('No rewrites are possible.')

		elif len(options) == 1:
			print('\nThere is only one option.')

			self.term = options[0]

		else:
			option = '#'

			while not option.isdecimal() or not (0 <= int(option) < len(options)):
				option = input(f'\nSelect one of the options (0-{len(options) -1}): ')

			self.term = options[int(option)]

	def print_context(self, ctx, lhs):
		"""Print the context"""
		var_name = f'<<PH>>:{lhs.getSort()}'
		var_term = self.module.parseTerm(var_name)

		ctx = ctx(var_term)

		return 'top' if ctx.isVariable() else \
		       str(ctx).replace(var_name, '\x1b[32m@\x1b[0m')

	def make_trs(self, out):
		"""Make a TRS file from the current module"""

		varset = set()

		for rl in self.module.getRules():
			lhs, rhs = rl.getLhs(), rl.getRhs()

			find_vars(lhs, varset)
			find_vars(rhs, varset)

		print_var = lambda v: f'{v.getVarName()}:{v.getSort()}'
		print('(VAR ' + ' '.join(map(print_var, varset)) + ')', file=out)
		print('(RULES', file=out)

		for rl in self.module.getRules():
			lhs, rhs = rl.getLhs(), rl.getRhs()
			print(f'\t{lhs.prettyPrint(0)} -> {rhs.prettyPrint(0)}', file=out)

		print(')', file=out)


if __name__ == '__main__':

	maude.init()

	# Command-line interface
	interface = InteractiveRewriter()

	interface.cmdloop()



import maude

maude.init(advise=False)
maude.load('smt')

cnv = maude.getModule('CONVERSION')
smt = maude.getModule('REAL-INTEGER')

def print_conversions(t):
	print(int(t))
	print(float(t))
	t.reduce()
	print(int(t))
	print(float(t))


print_conversions(cnv.parseTerm('12.34'))
print_conversions(cnv.parseTerm('567'))
print_conversions(cnv.parseTerm('-135'))
print_conversions(smt.parseTerm('89'))
print_conversions(smt.parseTerm('34/12'))
print_conversions(smt.parseTerm('false'))
print_conversions(cnv.parseTerm('s(N:Nat)'))
print_conversions(cnv.parseTerm('1 + 2'))
print_conversions(cnv.parseTerm('1.2 + 4.5'))



import maude
import sys

BAR         = '\033[1;31m|\033[0m'	# A bar in red
ARROW       = '\033[1;31m∨\033[0m'	# An arrow in red
CYCLE_BAR   = '\033[1;31m| |\033[0m'	# Two bars in red
CYCLE_ARR   = '\033[1;31m| ∨\033[0m'	# A bar and an arrow in red
CYCLE_END   = '\033[1;31m< ∨\033[0m'	# Two bars in red
SOLUTION    = '\033[1;32mX\033[0m'	# An X in green
EDGE_FMT    = '\033[3m\033[36m'		# Format for edges
RESET_FMT   = '\033[0m'			# Reset format

def print_smc_trans(trans):
	return {
		maude.StrategyRewriteGraph.RULE_APPLICATION: trans.getRule(),
		maude.StrategyRewriteGraph.OPAQUE_STRATEGY : trans.getStrategy(),
		maude.StrategyRewriteGraph.SOLUTION: ''
	}[trans.getType()]

def print_smc(sgraph, result):
	"""Print the result of model checking"""

	if result.holds:
		print('The property holds.')
	else:
		print('The property does not hold.')

		# Lead-in to the cycle
		for i in range(0, len(result.leadIn)):
			next_state = result.leadIn[i+1] if i+1 < len(result.leadIn) else result.cycle[0]
			trans = sgraph.getTransition(result.leadIn[i], next_state)

			print(BAR, sgraph.getStateTerm(result.leadIn[i]))
			print(ARROW, EDGE_FMT, print_smc_trans(trans), RESET_FMT)

		# Cycle
		for i in range(0, len(result.cycle)):
			next_state = result.cycle[i+1] if i+1 < len(result.cycle) else result.cycle[0]
			trans = sgraph.getTransition(result.cycle[i], next_state)

			if trans.getType() == maude.StrategyRewriteGraph.SOLUTION:
				print(SOLUTION, sgraph.getStateTerm(result.cycle[i]))
			else:
				print(CYCLE_BAR, sgraph.getStateTerm(result.cycle[i]))
				print(CYCLE_ARR, EDGE_FMT, print_smc_trans(trans), RESET_FMT)

		if (sgraph.getTransition(result.cycle[-1], result.cycle[0]).getType()
			!= maude.StrategyRewriteGraph.SOLUTION):
			print(CYCLE_END)


def print_mc(graph, result):
	if result.holds:
		print('The property holds.')
	else:
		print('The property does not hold.')

		# Path to the cycle
		for i in range(0, len(result.leadIn)):
			next_state = result.leadIn[i+1] if i+1 < len(result.leadIn) else result.cycle[0]

			print(BAR, graph.getStateTerm(result.leadIn[i]))
			print(ARROW, EDGE_FMT, graph.getRule(result.leadIn[i], next_state), RESET_FMT)

		# Cycle
		for i in range(0, len(result.cycle)):
			next_state = result.cycle[i+1] if i+1 < len(result.cycle) else result.cycle[0]
			rule = graph.getRule(result.cycle[i], next_state)

			if rule is None:
				print(SOLUTION, graph.getStateTerm(result.cycle[i]))
			else:
				print(CYCLE_BAR, graph.getStateTerm(result.cycle[i]))
				print(CYCLE_ARR, EDGE_FMT, rule, RESET_FMT)

		if graph.getRule(result.cycle[-1], result.cycle[0]) is not None:
			print(CYCLE_END)


if __name__ == '__main__':

	if len(sys.argv) < 4 or len(sys.argv) > 7:
		print('Pretty-printer for the Maude model checker results')
		print(sys.argv[0], '<file> <initial term> <LTL formula> [<strategy>]')
		sys.exit(0)

	filename, initial, formula = sys.argv[1:4]

	maude.init(advise=False)
	if not maude.load(filename):
		print('Error loading file.')
		sys.exit(1)

	mod = maude.getCurrentModule()

	# Parse the initial term
	t = mod.parseTerm(initial)

	if t is None:
		print('The initial term cannot be parsed.')
		sys.exit(2)

	# Parse the LTL formula
	f = mod.parseTerm(formula)

	if f is None:
		print('The LTL formula cannot be parsed.')
		sys.exit(3)

	# If model checking with strategy
	if len(sys.argv) > 4:
		# Parse the strategy expression
		s = mod.parseStrategy(sys.argv[4])

		if s is None:
			print('The strategy expression cannot be parsed.')
			sys.exit(4)

		print('Model checking', f, 'from', initial, 'using', s, 'in module', mod)

		sgraph = maude.StrategyRewriteGraph(t, s)
		result = sgraph.modelCheck(f)

		print_smc(sgraph, result)

	else:
		print('Model checking', f, 'from', initial, 'in module', mod)

		graph = maude.RewriteGraph(t)
		result = graph.modelCheck(f)

		print_mc(graph, result)



import os
import maude

maude.init(advise=False)

maude.input('''mod HOOK-TEST is
	protecting NAT .
	protecting STRING .
	sort Foo .

	op pair : Nat Nat -> Foo [special (
		id-hook SpecialHubSymbol (pair arg1 arg2)
		term-hook falseTerm (false)
		op-hook notSymbol (not_ : Bool ~> Bool)
	)] .

	op box : Nat -> Foo [special (
		id-hook SpecialHubSymbol (norewrite)
	)] .

	op getenv : String ~> String [special (
		id-hook SpecialHubSymbol
	)] .

	var N : Nat .

	rl [pred] : s(N) => N .
endm
''')

class ShowHook(maude.Hook):
	"""Show the data associated to the special operator"""

	def run(self, term, data):
		print(term)
		print(data.getData())
		print(data.getSymbol('notSymbol'))
		print(data.getTerm('falseTerm'))

		# Term are given unreduced to callbacks, which must reduce
		# them if to appear in the result (to avoid internal errors).
		for arg in term.arguments():
			arg.reduce()

		return None

class EnvHook(maude.Hook):
	"""Query environment variables (op getenv : String -> String)"""

	def run(self, term, data):
		# Environment variable
		name_term = next(term.arguments())
		name_term.reduce()

		variable = str(name_term)[1:-1]
		value = os.getenv(variable)

		# Current module
		module = term.symbol().getModule()

		return None if value is None else \
			module.parseTerm(f'"{os.getenv(variable)}"')

def make_hook(fn):
	"""Construct a hook from a function"""

	class TempHook(maude.Hook):
		def run(self, *args):
			return fn(*args)

	return TempHook()

def swap_pair(term, data):
	"""Swap the arguments of a pair"""

	args = list(term.arguments())
	args.reverse()

	return term.symbol().makeTerm(args)

def msg_none(term, data):
	"""Show a message"""
	print('--> msg_hook')

	# Even if the top symbol is not reduced (None is returned),
	# its arguments must be.
	for arg in term.arguments():
		arg.reduce()

	return None

# Instantiate the hook classes
printer    = ShowHook()
swaper     = make_hook(swap_pair)
no_rewrite = make_hook(lambda *args: None)
msg_hook   = make_hook(msg_none)
envhook    = EnvHook()

# Connect the special operators to the Python code
maude.connectEqHook('pair', printer)
maude.connectRlHook('pair', swaper)
maude.connectRlHook('norewrite', no_rewrite)
maude.connectEqHook('getenv', envhook)

# Equational and rule rewriting
m = maude.getCurrentModule()

t = m.parseTerm('pair(sd(5, 3), 1 + 1)')
t.rewrite(3)

print(t)

# Equational returing new terms
t = m.parseTerm('getenv("LANG")')
t.reduce()

print(t)

# Hook associated with a different name, to a lambda and doing nothing
t = m.parseTerm('box(4)')
t.rewrite()

print(t)


# Disconnecting hooks and default callbacks
print('-- connected')
m.parseTerm('pair(4, 5)').reduce()

print('-- disconnected')
maude.connectEqHook('pair', None)
m.parseTerm('pair(4, 5)').reduce()

print('-- connected + default')
maude.connectEqHook('pair', printer)
maude.connectEqHook(None, msg_hook)
m.parseTerm('pair(4, 5)').reduce()

print('-- default')
maude.connectEqHook('pair', None)
m.parseTerm('pair(4, 5)').reduce()

print('-- none')
maude.connectEqHook(None, None)
m.parseTerm('pair(4, 5)').reduce()



import maude

maude.init(advise=False)

nat = maude.getModule('NAT')
meta = maude.getModule('META-LEVEL')

zero = nat.parseTerm('0')
one  = nat.parseTerm('1')

t1 = meta.parseTerm("match '0.Zero s.t. nil")

s = nat.downStrategy(t1)

t2 = meta.upStrategy(s)

print(t1, '->', s, '->', t2)
print('Applied to', zero, '->', list(zero.srewrite(s)))
print('Applied to', one, '->', list(one.srewrite(s)))

#####

t = meta.parseTerm("'_+_['0.Zero, 'N:Nat]")

s1 = nat.downTerm(t)
s2 = meta.downTerm(t)

print(t, '->', s1, 'in NAT, ', s2, 'in META-LEVEL')

#####

t1 = nat.parseTerm("(1 + 2) * 3 + J:Nat")
t2 = nat.parseTerm("(1 + 2) * 3 + J:Nat")

s1 = meta.upTerm(t1)
s2 = meta.upTerm(t2)

print(t1, '->', s1)
print(t2, '->', s2)

print(s1 == s2)

#####

print(nat.upTerm(t1), nat.downTerm(t1))



#
# Benchmark the iterative reduction of a term using the Maude library and
# two alternative approaches
#

fib_maude = '''fmod FIBONACCI is
	protecting LIST{Nat} .

	op next : List{Nat} -> List{Nat} .

	var N M : Nat .
	var L   : List{Nat} .

	eq next(N M L) = (N + M) N M L .
endfm'''

sock_maude = r'''sload socket

mod FIBONACCI-SOCK is
	protecting FIBONACCI .
	protecting SOCKET .
	protecting CONVERSION .
	protecting (META-LEVEL + LEXICAL) * (
		op nil to eil, op __ to _#_,
		op append to $append,
		op head to $head,
		op tail to $tail,
		op last to $last,
		op front to $front,
		op occurs to $occurs,
		op reverse to $reverse,
		op $reverse to $$reverse,
		op size to $size,
		op $size to $$size
	) .

	op Fib : -> Cid [ctor] .
	op fib : -> Oid [ctor] .

	sort State .
	ops start wait-server wait-client active end : -> State [ctor] .

	op state:_ : State -> Attribute [ctor] .
	op l:_ : List{Nat} -> Attribute [ctor] .
	op ssock:_ : Oid -> Attribute [ctor] .
	op sock:_ : Oid -> Attribute [ctor] .

	vars S C : Oid .
	vars N M : Nat .
	var  T   : String .
	var  L   : List{Nat} .

	rl
		< fib : Fib | state: start >
	=>
		< fib : Fib | state: wait-server >
		createServerTcpSocket(socketManager, fib, 1234, 1)
	.

	rl
		< fib : Fib | state: wait-server >
		createdSocket(fib, socketManager, S)
	=>
		< fib : Fib | state: wait-client, ssock: S >
		acceptClient(S, fib)
	.

	rl
		< fib : Fib | state: wait-client, ssock: S >
		acceptedClient(fib, S, T, C)
	=>
		< fib : Fib | state: wait-client, ssock: S, sock: C >
		receive(C, fib)
	.

	crl
		< fib : Fib | state: wait-client, ssock: S, sock: C >
		received(fib, C, T)
	=>
		< fib : Fib | state: wait-client, ssock: S, sock: C, l: L >
		send(C, fib, "-\n")

	if L := downTerm(getTerm(metaParse(['FIBONACCI], tokenize(T), 'List`{Nat`})), nil) .

	rl
		< fib : Fib | state: wait-client, ssock: S, sock: C, l: L >
		sent(fib, C)
	=>
		< fib : Fib | state: active, ssock: S, sock: C, l: L >
		receive(C, fib)
	.

	rl
		< fib : Fib | state: active, ssock: S, sock: C, l: (N M L) >
		received(fib, C, "n\n")
	=>
		< fib : Fib | state: active, ssock: S, sock: C, l: ((N + M) N M L) >
		send(C, fib, string(N + M, 10) + "\n")
	.

	rl
		< fib : Fib | state: active, ssock: S, sock: C, l: L >
		sent(fib, C)
	=>
		< fib : Fib | state: active, ssock: S, sock: C, l: L >
		receive(C, fib)
	.

	rl
		< fib : Fib | state: active, ssock: S, sock: C, l: L >
		received(fib, C, "c\n")
	=>
		< fib : Fib | state: active, ssock: S, sock: C >
		closeSocket(C, fib)
	.

	rl
		< fib : Fib | state: active, ssock: S, sock: C >
		closedSocket(fib, C, T)
	=>
		< fib : Fib | state: active, ssock: S >
		closeSocket(S, fib)
	.

	rl
		< fib : Fib | state: active, ssock: S >
		closedSocket(fib, S, T)
	=>
		< fib : Fib | state: end >
	.
endm
'''


def spiral_maude():
	import maude

	maude.init()
	maude.input(fib_maude)
	cm = maude.getCurrentModule()

	x = cm.parseTerm(f'1 0')

	nat_kind = cm.findSort('Nat').kind()

	next_symb = cm.findSymbol('next', (nat_kind, ), nat_kind)

	while True:
		yield int(str(next(x.arguments())))
		x = next_symb(x)
		x.reduce()


def maude_reduce(p, term):
	p.stdin.write(f'red {term} .\n'.encode('ascii'))
	p.stdin.flush()

	answer = p.stdout.readline()

	return answer[answer.index(b':') + 1:].strip().decode('ascii')


def spiral_popen():
	import os
	import subprocess

	p = subprocess.Popen([os.getenv('MAUDE_BIN') or 'maude', '-no-banner', '-no-wrap'],
	                     stdin=subprocess.PIPE, stdout=subprocess.PIPE)

	p.stdin.write(fib_maude.encode('ascii'))
	p.stdin.write(b'\nset show command off . set show stats off .\n')
	p.stdin.flush()

	x = f'1 0'

	while True:
		yield int(x.split()[0])
		x = maude_reduce(p, f'next({x})')


def spiral_socket():
	import os
	import socket
	import subprocess

	p = subprocess.Popen([os.getenv('MAUDE_BIN') or 'maude', '-no-banner'],
	                     stdin=subprocess.PIPE, stdout=subprocess.DEVNULL)
	p.stdin.write(fib_maude.encode('ascii') + b'\n' + sock_maude.encode('ascii')
	              + b'\nerew <> < fib : Fib | state: start > .\nquit .\n')
	p.stdin.flush()

	# Connect to the port opened by Maude
	raw_sock = None

	while raw_sock is None:
		try:
			raw_sock = socket.create_connection(('127.0.0.1', 1234))

		except ConnectionRefusedError:
			pass

	x = '1 0\n'

	with raw_sock.makefile('rw') as sock:
		sock.write(x)
		sock.flush()
		sock.readline()

		while True:
			yield int(x.split()[0])
			sock.write('n\n')
			sock.flush()
			x = sock.readline()

		# sock.write('c\n')
		# sock.flush()

	# raw_sock.close()
	# p.wait()


if __name__ == '__main__':
	import argparse
	import itertools

	parser = argparse.ArgumentParser(description='Spiral generator')
	parser.add_argument('--draw', help='Draw the spiral', action='store_true')
	parser.add_argument('--eps', help='Epsilon for the module', type=float, default=1e-3)
	parser.add_argument('imp', help='Implementation', choices=['lib', 'io', 'socket', 'py'], default='lib')
	parser.add_argument('index', help='Index of the series to compute', type=int)

	args = parser.parse_args()

	imp = {'lib': spiral_maude, 'io': spiral_popen, 'socket': spiral_socket}[args.imp]

	for f in itertools.islice(imp(), args.index):
		print(f)



import maude
import sys

def printElements(name, elems):
	if len(elems) == 0:
		return

	print('<h2>', name, '</h2>')
	print('<ul>')

	for elem in elems:
		print('<li>', elem)

	print('</ul>')

MODULE_TYPE_NAMES = {
	maude.Module.FUNCTIONAL_MODULE: "fmod",
	maude.Module.SYSTEM_MODULE: "smod",
	maude.Module.STRATEGY_MODULE: "smod",
	maude.Module.FUNCTIONAL_THEORY: "fth",
	maude.Module.SYSTEM_THEORY: "th",
	maude.Module.STRATEGY_THEORY: "sth"
}


if __name__ == "__main__":

	maude.init(advise=False)

	# Get the current module if a module name has not been
	# given in the first argument
	if len(sys.argv) >= 2:
		m = maude.getModule(sys.argv[1])
	else:
		m = maude.getCurrentModule()

	print("""<!DOCTYPE HTML>
	<html>
	<head>
		<meta charset="utf-8" />
	</head>
	<body>
	""")

	print("<h1>", MODULE_TYPE_NAMES[m.getModuleType()], m, "</h1>")

	printElements('Sorts', m.getSorts())
	printElements('Operators', m.getSymbols())
	printElements('Membership axioms', m.getMembershipAxioms())
	printElements('Equations', m.getEquations())
	printElements('Rules', m.getRules())
	printElements('Strategies', m.getStrategies())
	printElements('Strategy definitions', m.getStrategyDefinitions())

	print("""</body>
</html>""")



import maude
import itertools

maude.init(advise=False)

##### From Maude 3.1 manual, §13.4

maude.input('''fmod UNIFICATION-EX1 is
	protecting NAT .
	op f : Nat Nat -> Nat .
	op f : NzNat Nat -> NzNat .
	op f : Nat NzNat -> NzNat .
endfm''')

uex1 = maude.getModule('UNIFICATION-EX1')

uex1_t1 = uex1.parseTerm('f(X:Nat, Y:Nat) ^ B:NzNat')
uex1_t2 = uex1.parseTerm('A:NzNat ^ f(Y:Nat, Z:Nat)')

print('Unifiers for', uex1_t1, 'and', uex1_t2)

for unifier in uex1.unify([(uex1_t1, uex1_t2)]):
	print('Unifier', unifier)
	print('X =', unifier.find('X'))
	print('T =', unifier.find('T'))
	print('B:NzNat =', unifier.find('B', uex1.findSort('NzNat')))
	print('X:NzNat =', unifier.find('X', uex1.findSort('NzNat')))

	print('σ({}) = {}'.format(uex1_t1, unifier.instantiate(uex1_t1)))
	print('σ(3) =', unifier.instantiate(uex1.parseTerm('3')))

#####

uex1_p1 = uex1.parseTerm('f(X:Nat, Y:NzNat)'), uex1.parseTerm('f(Z:NzNat, U:Nat)')
uex1_p2 = uex1.parseTerm('V:NzNat'), uex1.parseTerm('f(X:Nat, U:Nat)')

print('Unifier for equations', uex1_p1, 'and', uex1_p2)

for unifier in uex1.unify([uex1_p1, uex1_p2]):
	print(unifier)

#####

maude.input('''fmod UNIFICATION-EX3 is
	protecting NAT .
	op f : Nat Nat -> Nat [idem] .
endfm''')

uex3 = maude.getModule('UNIFICATION-EX3')

print('Unsupported unification')

result = uex3.unify([(uex3.parseTerm('f(f(X:Nat, Y:Nat), Z:Nat)'), uex3.parseTerm('f(A:Nat, B:Nat)'))])


##### From Maude 3.1 manual, §13.4.4

maude.input('''mod UNIF-VENDING-MACHINE is
	sorts Coin Item Marking Money State .
	subsort Coin < Money .
	op empty : -> Money .
	op __ : Money Money -> Money [assoc comm id: empty] .
	subsort Money Item < Marking .
	op __ : Marking Marking -> Marking [assoc comm id: empty] .
	op <_> : Marking -> State .
	ops $ q : -> Coin .
	ops a c : -> Item .
	var M : Marking .
	rl [buy-c] : < M $ > => < M c > .
	rl [buy-a] : < M $ > => < M a q > .
	eq [change]: q q q q = $ .
endm''')

uvmachine = maude.getModule('UNIF-VENDING-MACHINE')
uvm_p1    = uvmachine.parseTerm('< q q X:Marking >'), uvmachine.parseTerm('< $ Y:Marking >')

print('Unify', uvm_p1)

for unifier in uvmachine.unify([uvm_p1]):
	print(unifier)

print('Irredudant unify', uvm_p1)

for unifier in uvmachine.unify([uvm_p1], True):
	print(unifier)

print('Variant unify', uvm_p1)

for unifier in uvmachine.variant_unify([uvm_p1]):
	print(unifier)

##### From Maude 3.1 manual, §14.3

maude.input('''mod VARIANT-VENDING-MACHINE is
	sorts Coin Item Marking Money State .
	subsort Coin < Money .
	op empty : -> Money .
	op __ : Money Money -> Money [assoc comm id: empty] .
	subsort Money Item < Marking .
	op __ : Marking Marking -> Marking [assoc comm id: empty] .
	op <_> : Marking -> State .
	ops $ q : -> Coin .
	ops a c : -> Item .
	var M : Marking .
	rl [buy-c] : < M $ > => < M c > .
	rl [buy-a] : < M $ > => < M a q > .
	eq [change] : q q q q M = $ M [variant] .
endm''')

vvmachine = maude.getModule('VARIANT-VENDING-MACHINE')
vvm_p1    = vvmachine.parseTerm('< q q X:Marking >'), vvmachine.parseTerm('< $ Y:Marking >')

##### From §14.8

print('Variant unify', vvm_p1)

for unifier in vvmachine.variant_unify([vvm_p1]):
	print(unifier)

print('Filtered variant unify', vvm_p1)

for unifier in vvmachine.variant_unify([vvm_p1], filtered=True):
	print(unifier)

##### From §14.9

print('Variant unify', vvm_p1, 'with constraints')

irreducible = list(map(vvmachine.parseTerm, ['q q X:Marking', 'q X:Marking', 'X:Marking']))

for unifier in vvmachine.variant_unify([vvm_p1], irreducible):
	print(unifier)

##### From §14.12

print('Variant match', vvm_p1)

for matcher in vvmachine.variant_match([vvm_p1]):
	print(matcher)

print('Variant match', vvm_p1, 'with constraints')

for matcher in vvmachine.variant_match([vvm_p1], irreducible):
	print(matcher)


##### From Maude manual 3.1, §14.1 and 14.10

maude.input('''fmod NAT-VARIANT is
	sort Nat .
	op 0 : -> Nat [ctor] .
	op s : Nat -> Nat [ctor] .
	op _+_ : Nat Nat -> Nat .
	vars X Y : Nat .
	eq [base] : 0 + Y = Y [variant] .
	eq [ind] : s(X) + Y = s(X + Y) [variant] .
endfm''')

natv  = maude.getModule('NAT-VARIANT')
onex  = natv.parseTerm('s(0) + X:Nat')
xone  = natv.parseTerm('X:Nat + s(0)')
three = natv.parseTerm('s(s(s(0)))')

print('Variant unify', (onex, three))

for unifier in natv.variant_unify([(onex, three)]):
	print(unifier)

print('Variant unify', (onex, three))

for unifier in itertools.islice(natv.variant_unify([(xone, three)]), 1):
	print(unifier)



import maude
import os.path

def exploreAndGraph(graph, stateNr):
	print(stateNr, '[label="' + str(graph.getStateTerm(stateNr)) + '"];')

	index = 0
	nextState = graph.getNextState(stateNr, index)

	while nextState >= 0:
		print(stateNr, '->', nextState, ';')

		if nextState > stateNr:
			exploreAndGraph(graph, nextState)

		index = index + 1
		nextState = graph.getNextState(stateNr, index)

maude.init(advise=False)
maude.load(os.path.join(os.path.dirname(__file__), '..', 'example.maude'))

example = maude.getModule('EXAMPLE')
initial = example.parseTerm('f(a, a)')
graph = maude.RewriteGraph(initial)

print('digraph {')
exploreAndGraph(graph, 0)
print('}')



#
# Test for the lifetime of modules
#

import maude
import gc


class ModuleLifeTest:
	# Based on https://github.com/fadoss/maude-bindings/issues/3
	MODULE_TEMPLATE = r'''smod ONE is
		protecting NAT .
		op value : -> Nat .
		eq value = {value} .
		mb false : Bool .
		crl 1 => 0 if 1 = 1 /\ 1 : Nat /\ 1 => 1 .
		strat st @ Nat .
		sd st := idle .
	endsm'''

	def __init__(self):
		self.index = 0
		maude.input(self.MODULE_TEMPLATE.format(value=0))

	def run(self, name, first, second):
		"""Run an operation on an object obtained from a garbage collected module"""

		# Make the first function generate an object in the module ONE
		obj = first(maude.getCurrentModule())
		# Input a module overwriting the module ONE
		maude.input(self.MODULE_TEMPLATE.format(value=self.index))
		self.index += 1
		# Force garbage collection (if not done yet)
		gc.collect()
		# Make the second function use the object referring to the garbage collected module
		second(obj)

		print(f'[{name}]')


maude.init(advise=False)
mlt = ModuleLifeTest()

# Basic objects
mlt.run('term', lambda m: m.parseTerm('value'), lambda t: t.reduce())
mlt.run('term', lambda m: m.parseTerm('value'), lambda t: print(t))
mlt.run('sort-findSort', lambda m: m.findSort('NzNat'), lambda s: print(s))
mlt.run('sort-getSorts', lambda m: m.getSorts()[0], lambda s: print(s))
mlt.run('sort-MbAx', lambda m: m.getMembershipAxioms()[0].getSort(), lambda s: print(s))
mlt.run('sort-getRangeSort', lambda m: m.getSymbols()[0].getRangeSort(), lambda s: print(s))
mlt.run('sort-getSubjectSort', lambda m: m.parseTerm('1 + 2').getSort(), lambda s: print(s))
mlt.run('sort-getStrategies', lambda m: m.getStrategies()[0].getSubjectSort(), lambda s: print(s))
mlt.run('kind', lambda m: m.findSort('NzNat').kind(), lambda s: print(s))
mlt.run('symbol-getSymbols', lambda m: m.getSymbols()[0], lambda s: print(s))
mlt.run('equation-getEquations', lambda m: m.getEquations()[0], lambda e: print(e))
mlt.run('MbAx-getMbAx', lambda m: m.getMembershipAxioms()[0], lambda mb: print(mb))
mlt.run('rule-getRules', lambda m: m.getRules()[0], lambda r: print(r))
mlt.run('strategy-getStrategies', lambda m: m.getStrategies()[0], lambda r: print(r))
mlt.run('sds-getSds', lambda m: m.getStrategyDefinitions()[0], lambda d: print(d))
# mlt.run('strategy', lambda m: m.parseStrategy('st'), lambda s: print(s))

# Solution iterators
mlt.run('match', lambda m: m.parseTerm('value').match(m.parseTerm('N:Nat')), lambda i: next(i))
mlt.run('search', lambda m: m.parseTerm('value').search(maude.ANY_STEPS, m.parseTerm('N:Nat')), lambda i: next(i))
mlt.run('srewrite', lambda m: m.parseTerm('value').srewrite(m.parseStrategy('idle')), lambda i: next(i))
mlt.run('ssearch', lambda m: m.parseTerm('value').search(maude.ANY_STEPS, m.parseTerm('N:Nat'), strategy=m.parseStrategy('idle')), lambda i: next(i))
mlt.run('vu_narrow', lambda m: m.parseTerm('value').vu_narrow(maude.ANY_STEPS, m.parseTerm('N:Nat')), lambda i: next(i))
mlt.run('get_variants', lambda m: m.parseTerm('1 + N:Nat').get_variants(), lambda i: next(i))
mlt.run('unify', lambda m: m.unify([(m.parseTerm('1 + N:Nat'), m.parseTerm('M:Nat + 2'))]), lambda i: next(i))
mlt.run('variant_unify', lambda m: m.variant_unify([(m.parseTerm('1 + N:Nat'), m.parseTerm('M:Nat + 2'))]), lambda i: next(i))

# Vectors
mlt.run('getSorts', lambda m: m.getSorts(), lambda sv: print(sv[0]))
mlt.run('getSymbols', lambda m: m.getSymbols(), lambda sv: print(sv[0]))
mlt.run('getKinds', lambda m: m.getKinds(), lambda kv: print(kv[0]))
mlt.run('getMembershipAxioms', lambda m: m.getMembershipAxioms(), lambda mv: print(mv[0]))
mlt.run('getEquations', lambda m: m.getEquations(), lambda ev: print(ev[0]))
mlt.run('getRules', lambda m: m.getRules(), lambda rv: print(rv[0]))
mlt.run('getStrategies', lambda m: m.getStrategies(), lambda sv: print(sv[0]))
mlt.run('getStrategyDefinitions', lambda m: m.getStrategyDefinitions(), lambda dv: print(dv[0]))
mlt.run('getOpDeclarations', lambda m: m.parseTerm('1 + 2').symbol().getOpDeclarations(), lambda dv: print(dv[0].getDomainAndRange()[0]))
mlt.run('getSupersorts', lambda m: m.getSorts()[0].getSupersorts(), lambda sv: print(sv[0]))
mlt.run('getDomainAndRange', lambda m: m.parseTerm('1 + 2').symbol().getOpDeclarations()[0].getDomainAndRange(), lambda sv: print(sv[0]))
#mlt.run('getSubsorts', lambda m: m.getSorts()[0].getSubsorts(), lambda sv: print(len(sv)))
#mlt.run('getDomain', lambda m: m.getStrategies()[0].getDomain(), lambda sv: print(len(sv)))



import maude

maude.init(advise=False)

maude.load('../example.maude')
maude.load('tests/example.maude')
maude.input('fmod TESTS is endfm')
maude.input('fmod TESTS is endfm')
maude.load('../example.maude')
maude.input('fmod TESTS is endfm')


m = maude.getModule('NAT')
seven = m.parseTerm('7')

print(seven)
maude.input('set print number off .')
print(seven)



import maude
import os.path

maude.init(advise=False)
maude.load(os.path.join(os.path.dirname(__file__), '..', 'example.maude'))

example = maude.getCurrentModule()
nat = maude.getModule('NAT')

#####

t = nat.parseTerm('1 + (3 * 5 + 7) * 11 + 13')
t0 = t.copy()
t.reduce()
print(t0, '=', t)

#####

ans = example.parseTerm('f(a, b)')
ans.rewrite()
print(ans)

#####

ans = example.parseTerm('f(a, a)')
ans.frewrite(1)
print(ans)

#####

fbb = example.parseTerm('f(b, b)')
ans, nrew = fbb.erewrite()
print(fbb, '->', ans, 'in', nrew, 'rewrites')

#####

ans = example.parseTerm('f(a, b)')

for sol, nrew in ans.srewrite(example.parseStrategy('swap *')):
	print(sol, 'in', nrew, 'rewrites')

#####

initial = example.parseTerm('f(a, a)')
pattern = example.parseTerm('f(c, X:Symbol)')

for sol, subs, path, nrew in initial.search(maude.ANY_STEPS, pattern):
	print(sol, 'with', subs, 'by', path())

s = example.parseStrategy('ab ; bc ; ab')

for sol, subs, path, nexts, nrew in initial.search(maude.ANY_STEPS, pattern, s):
	print(sol, 'with', subs, 'by', path(), 'before applying', nexts)



import maude
import itertools

maude.init(advise=False)

def print_solution(*args):
	print(('              state = {}\n'
	       '  acc. substitution = {}\n'
	       '    variant unifier = {}').format(*args))

##### From Maude 3.0 manual, §15.6

maude.input(r'''mod NARROWING-VENDING-MACHINE is
	sorts Coin Item Marking Money State .
	subsort Coin < Money .
	op empty : -> Money .
	op __ : Money Money -> Money [assoc comm id: empty] .
	subsort Money Item < Marking .
	op __ : Marking Marking -> Marking [assoc comm id: empty] .
	op <_> : Marking -> State .
	ops $ q : -> Coin .
	ops a c : -> Item .
	var M : Marking .
	rl [buy-c] : < M $ > => < M c > [narrowing] .
	rl [buy-a] : < M $ > => < M a q > [narrowing] .
	eq [change] : q q q q M = $ M [variant] .
endm''')

nvmach = maude.getModule('NARROWING-VENDING-MACHINE')

nvmach_initial1 = nvmach.parseTerm('< M:Money >')
nvmach_target1  = nvmach.parseTerm('< a c >')
nvmach_initial2 = nvmach.parseTerm('< C1:Coin C2:Coin C3:Coin C4:Coin >')
nvmach_target2  = nvmach.parseTerm('< M:Money a c >')
nvmach_initial3 = nvmach.parseTerm('< $ q q q M1:Money >')
nvmach_target3  = nvmach.parseTerm('< a c M2:Money >')

print(nvmach_initial1, '=>*', nvmach_target1)

for term, subs, unifier in itertools.islice(nvmach_initial1.vu_narrow(maude.ANY_STEPS, nvmach_target1), 1):
	print_solution(term, subs, unifier)

print(nvmach_initial1, '=>!', nvmach_target1, 'with depth bound to 5')

for term, subs, unifier in nvmach_initial1.vu_narrow(maude.NORMAL_FORM, nvmach_target1, 5):
	print_solution(term, subs, unifier)

print(nvmach_initial2, '=>!', nvmach_target2, 'with depth bound to 10')

for term, subs, unifier in nvmach_initial2.vu_narrow(maude.NORMAL_FORM, nvmach_target2, 10):
	print_solution(term, subs, unifier)

print(nvmach_initial1, '=>*', nvmach_target1, 'with folding')

for solution in itertools.islice(nvmach_initial1.vu_narrow(maude.ANY_STEPS, nvmach_target1, -1, True), 1):
	print_solution(*solution)

print(nvmach_initial3, '=>*', nvmach_target3, 'with depth bound to 2')

for solution in nvmach_initial3.vu_narrow(maude.ANY_STEPS, nvmach_target3, depth=2):
	print_solution(*solution)

print(nvmach_initial3, '=>*', nvmach_target3, 'with depth bound to 5 and filter')

for solution in nvmach_initial3.vu_narrow(maude.ANY_STEPS, nvmach_target3, depth=5, flags=maude.FILTER):
	print_solution(*solution)

print(nvmach_initial3, '=>*', nvmach_target3, 'with depth bound to 5 and delay')

for solution in nvmach_initial3.vu_narrow(maude.ANY_STEPS, nvmach_target3, depth=5, flags=maude.DELAY | maude.FILTER):
	print_solution(*solution)

##### From §15.7

maude.input(r'''mod FOLDING-NARROWING-VENDING-MACHINE is
	sorts Coin Item Marking Money State .
	subsort Coin < Money .
	op empty : -> Money .
	op __ : Money Money -> Money [assoc comm id: empty] .
	subsort Money Item < Marking .
	op __ : Marking Marking -> Marking [assoc comm id: empty] .
	op <_> : Marking -> State .
	ops $ q : -> Coin .
	ops a c : -> Item .
	var M : Marking .
	rl [buy-c] : < M $ c > => < M > [narrowing] .
	rl [buy-a] : < M $ a > => < M q > [narrowing] .
	eq [change] : q q q q M = $ M [variant] .
endm''')

fnvmach = maude.getModule('FOLDING-NARROWING-VENDING-MACHINE')

fnvmach_initial  = fnvmach.parseTerm('< M:Marking a c >')
fnvmach_initial2 = fnvmach.parseTerm('< M:Money a c >')
fnvmach_target   = fnvmach.parseTerm('< empty >')

print(fnvmach_initial, '=>*', fnvmach_target, 'with folding')

for solution in fnvmach_initial.vu_narrow(maude.ANY_STEPS, fnvmach_target, -1, maude.FOLD):
	print_solution(*solution)

print(fnvmach_initial2, '=>*', fnvmach_target)

for solution in fnvmach_initial2.vu_narrow(maude.ANY_STEPS, fnvmach_target):
	print_solution(*solution)


##### New commands from Maude 3.5

vu_narrow_it = nvmach_initial1.vu_narrow(maude.ANY_STEPS, nvmach_target1, 3, maude.FOLD)

list(vu_narrow_it) # read the sequence

print('show most general states', nvmach_initial1, '=>*', nvmach_target1)

for state in vu_narrow_it.getMostGeneralStates():
	print(state)

print('show frontier states', nvmach_initial1, '=>*', nvmach_target1)

for state in vu_narrow_it.getFrontierStates():
	print(state)

vu_narrow_it = nvmach_initial2.vu_narrow(maude.NORMAL_FORM, nvmach_target2)

list(vu_narrow_it) # read the sequence

print('show frontier states', nvmach_initial2, '=>!', nvmach_target2)

for state in vu_narrow_it.getFrontierStates():
	print(state)

print('vu-narrow from multiple initial terms')

vu_narrow_it = fnvmach.vu_narrow([
	fnvmach.parseTerm('< M1:Marking a c >'),
	fnvmach.parseTerm('< M2:Marking a a >'),
	fnvmach.parseTerm('< M3:Marking c c >')], maude.ANY_STEPS, fnvmach.parseTerm('< empty >'), flags=maude.FOLD)

for solution in vu_narrow_it:
	print_solution(*solution)

for state in vu_narrow_it.getFrontierStates():
	print(state)

for state in vu_narrow_it.getMostGeneralStates():
	print(state)



import maude

maude.init(advise=False)

mod = maude.getModule('NAT')
natk = mod.findSort('Nat').kind()

splus  = mod.findSymbol('_+_', [natk, natk], natk)
stimes = mod.findSymbol('_*_', [natk, natk], natk)

onetwo = [mod.parseTerm('1'), mod.parseTerm('2')]
three  = mod.parseTerm('3')

# Constructs 4 + (3 * (1 + 2))
expr = splus.makeTerm((mod.parseTerm('4'),
                      stimes.makeTerm([three,
                                      splus.makeTerm(onetwo)])))

print(expr.getSort(), ':', expr)
expr.reduce()
print(expr)

# Constructs a variable by parsing
var = mod.parseTerm('N:Nat')

print(var, var.isVariable(), var.getVarName())
print(expr, expr.isVariable(), expr.getVarName())



#
# Benchmark reducing a constant with the Maude library and two other approaches
#

import os
import socket
import subprocess
import time
import timeit

import maude


def reduce_io():
	"""Reduce 0 by input/output interaction with a Maude interpreter process"""

	p.stdin.write(b'red 0 .\n')
	p.stdin.flush()

	answer = p.stdout.readline()

	return answer[answer.index(b':') + 1:].strip().decode('ascii')


def reduce_lib():
	"""Reduce 0 using the maude language bindings"""

	t = m.parseTerm('0')
	t.reduce()
	return str(t)


def reduce_socket():
	"""Reduce 0 using a Maude-based reducer server"""

	sock.write('0\n')
	sock.flush()
	return sock.readline().strip()


if __name__ == '__main__':

	# Maude binary to be used
	maude_binary = os.getenv('MAUDE_BIN')

	if not maude_binary:
		maude_binary = 'maude'

	# The maude library
	maude.init()
	m = maude.getCurrentModule()

	print('lib', timeit.timeit('reduce_lib()', setup='from __main__ import reduce_lib'))

	# The I/O approach
	p = subprocess.Popen([maude_binary, '-no-banner', '-no-wrap'],
	                     stdin=subprocess.PIPE, stdout=subprocess.PIPE)

	p.stdin.write(b'\nset show command off . set show stats off .\n')
	p.stdin.flush()

	print('io', timeit.timeit('reduce_io()', setup='from __main__ import reduce_io'))

	# The Maude-based server approach
	p.stdin.write(b'\nsload bm-reduce\nquit .\n')
	p.stdin.flush()

	time.sleep(2)  # Wait for the server to be started

	raw_sock = socket.create_connection(('127.0.0.1', 1234))
	sock = raw_sock.makefile('rw')

	print('socket', timeit.timeit('reduce_socket()', setup='from __main__ import reduce_socket'))

	sock.write('c\n')
	sock.close()
	raw_sock.close()
	p.wait()



import maude, sys
from tkinter import filedialog, messagebox
import tkinter as tk
import tkinter.ttk as ttk

class ReduceWindow:
	def __init__(self):
		self.window = tk.Tk()
		self.window.title('Maude bindings example')

		self.tbox = tk.Entry(self.window)
		self.tbox.grid(column=0, row=0)
		self.btn_tree = tk.Button(self.window, text='Tree', command=self.fillTree)
		self.btn_tree.grid(column=1, row=0)
		self.btn_red = tk.Button(self.window, text='Reduce', command=lambda: self.fillTree(True))
		self.btn_red.grid(column=2, row=0)
		self.tree = ttk.Treeview(self.window, columns=('sort'))
		self.tree.grid(column=0, row=1, columnspan=3)

		self.tree.heading('#0', text='Subterm')
		self.tree.heading('sort', text='Sort')

	def fillTree(self, doReduce=False):
		term = self.module.parseTerm(self.tbox.get())
		if term is None:
			messagebox.showerror('Maude bindings example', 'No parse for that term')
		else:
			if doReduce:
				term.reduce()
				self.tbox.delete(0, tk.END)
				self.tbox.insert(0, str(term))

			self.tree.delete(*self.tree.get_children())

			parent = self.tree.insert('', tk.END, text=str(term), values=str(term.getSort()))
			self.add_subterms(term, parent)

	def add_subterms(self, term, parent):
		for subterm in term.arguments():
			entry = self.tree.insert(parent, tk.END, text=str(subterm), values=str(subterm.getSort()))
			self.add_subterms(subterm, entry)

	def run(self):
		filename = filedialog.askopenfilename(title='Select Maude file',
				filetypes=[('Maude files', '*.maude')])

		if len(filename) == 0:
			sys.exit(1)

		maude.load(filename)
		self.module = maude.getCurrentModule()
		messagebox.showinfo('Maude bindings example',
			'Selected module is ' + str(self.module))

		self.window.mainloop()


if __name__ == '__main__':
	maude.init(advise=False)

	window = ReduceWindow()
	window.run()



import maude
import os.path

maude.init(advise=False)
maude.load(os.path.join(os.path.dirname(__file__), '..', 'example.maude'))

m = maude.getCurrentModule()

print('Using', m, 'module')

#####

pattern = m.parseTerm('f(X:Symbol, Y:Symbol)')
t = m.parseTerm('f(a, g(b))')

print(pattern, '<=?', t)

for match, _ in t.match(pattern):
	print(match)

#####

pattern = m.parseTerm('X:Symbol ; Y:Symbol')
t = m.parseTerm('b ; c')

print(pattern, '<=?', t)

for match, _ in t.match(pattern):
	print(match)

#####

pattern = m.parseTerm('X:Symbol ; Y:Symbol')
t = m.parseTerm('b ; c')

print(pattern, '<=?', t)

for match, _ in t.match(pattern, [maude.EqualityCondition(m.parseTerm('X:Symbol'), m.parseTerm('Y:Symbol'))]):
	print(match)

#####

cond = [maude.EqualityCondition(m.parseTerm('X:Symbol'), m.parseTerm('Y:Symbol'))]

pattern = m.parseTerm('f(X:Symbol, g(Y:Symbol))')
t = m.parseTerm('f(a, g(a))')

print(pattern, '<=?', t)

for match, _ in t.match(pattern, cond):
	print(match)

#####

pattern = m.parseTerm('X:Symbol Y:Symbol')
t = m.parseTerm('a b c g(b) f(a, b)')

print(pattern, '<=?', t)

for match, _ in t.match(pattern, maude.Term.NO_CONDITION, withExtension=True):
	print(match, 'inside', match.matchedPortion())

#####

pattern = m.parseTerm('f(X:Symbol, g(Y:Symbol))')
t = m.parseTerm('f(a, g(b))')

for match, _ in t.match(pattern):
	print('Substitution', match)
	print('X =', match.find('X'))
	print('Y =', match.find('Y'))
	print('X:Symbol =', match.find('X', m.findSort('Symbol')))
	print('Y:Nat =', match.find('Y', m.findSort('Nat')))
	print('σ(c) =', match.instantiate(m.parseTerm('c')))
	print('σ(f(g(Y), g(X))) =', match.instantiate(m.parseTerm('f(g(Y:Symbol), g(X:Symbol))')))

#####

pattern = m.parseTerm('f(X:Symbol, Y:Symbol)')
t = m.parseTerm('f(f(g(a), a), f(b, c))')
h = m.parseTerm('%hole:Symbol')

for match, ctx in t.match(pattern, maxDepth=maude.UNBOUNDED):
	print(match, '---', ctx(h), '---', ctx(m.parseTerm('a')))

for match, ctx in t.match(pattern, minDepth=1, maxDepth=2):
	print(match)



import maude
import itertools

maude.init(advise=False)

def print_variant(term, subs):
	print('{:>28} -- {}'.format(str(term), subs))

##### From Maude 3.0 manual, §14.1 and 14.3

maude.input('''fmod EXCLUSIVE-OR is
	sorts Nat NatSet .
	op 0 : -> Nat [ctor] .
	op s : Nat -> Nat [ctor] .

	subsort Nat < NatSet .
	op mt : -> NatSet [ctor] .
	op _*_ : NatSet NatSet -> NatSet [ctor assoc comm] .

	vars X Y Z : [NatSet] .
	eq [idem] :
	X * X = mt [variant] .
	eq [idem-Coh] : X * X * Z = Z [variant] .
	eq [id] :
	X * mt = X [variant] .
endfm''')

xor  = maude.getModule('EXCLUSIVE-OR')
xory = xor.parseTerm('X * Y')

print('Irredundant variants of', xory)

for term, subs in xory.get_variants(True):
	print_variant(term, subs)

print('\nIrredundant variants of', xory, 'such that', xory, 'irreducible')

for term, subs in xory.get_variants(True, [xory]):
	print_variant(term, subs)


##### From Maude 3.0 manual, §14.1 and 14.5

maude.input('''fmod NAT-VARIANT is
	sort Nat .
	op 0 : -> Nat [ctor] .
	op s : Nat -> Nat [ctor] .
	op _+_ : Nat Nat -> Nat .
	vars X Y : Nat .
	eq [base] : 0 + Y = Y [variant] .
	eq [ind] : s(X) + Y = s(X + Y) [variant] .
endfm''')

natv = maude.getModule('NAT-VARIANT')
onex = natv.parseTerm('s(0) + X:Nat')
xone = natv.parseTerm('X:Nat + s(0)')

print('\nVariants of', onex)

for term, subs in onex.get_variants():
	print_variant(term, subs)

print('\n10 variants of', xone)

for term, subs in itertools.islice(xone.get_variants(), 10):
	print_variant(term, subs)
